# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Utility functions for Astro-SCRAPPY

These include fast implementations for calculating the median,
median filters, and other image operations.
"""

from .median_utils import *  # noqa
from .image_utils import *  # noqa

__all__ = ['median', 'optmed3', 'optmed5', 'optmed7', 'optmed9',  # noqa
           'optmed25', 'medfilt3', 'medfilt5', 'medfilt7',
           'sepmedfilt3', 'sepmedfilt5', 'sepmedfilt7', 'sepmedfilt9',
           'subsample', 'rebin', 'convolve', 'laplaceconvolve',
           'dilate3', 'dilate5']



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest

from . import fake_data
from ..astroscrappy import detect_cosmics


@pytest.fixture
def testdata():
    # Get fake data to work on
    imdata, crmask = fake_data.make_fake_data()
    return imdata, crmask


@pytest.mark.xfail
def test_median_clean(testdata):
    imdata, crmask = testdata
    # Because our image only contains single cosmics, turn off
    # neighbor detection. Also, our cosmic rays are high enough
    # contrast that we can turn our detection threshold up.
    _mask, clean = detect_cosmics(imdata, readnoise=10., gain=1.0,
                                  sigclip=6, sigfrac=1.0, cleantype='median')

    assert (clean[crmask] != imdata[crmask]).sum() == crmask.sum()

    # Run it again on the clean data. We shouldn't find any new cosmic rays
    _mask2, _clean2 = detect_cosmics(clean, readnoise=10., gain=1.0,
                                     sigclip=6, sigfrac=1.0, cleantype='median')
    assert _mask2.sum() == 0


def test_medmask_clean(testdata):
    imdata, crmask = testdata
    # Because our image only contains single cosmics, turn off
    # neighbor detection. Also, our cosmic rays are high enough
    # contrast that we can turn our detection threshold up.
    _mask, clean = detect_cosmics(imdata, readnoise=10., gain=1.0,
                                  sigclip=6, sigfrac=1.0, cleantype='medmask')

    assert (clean[crmask] != imdata[crmask]).sum() == crmask.sum()

    # Run it again on the clean data. We shouldn't find any new cosmic rays
    _mask2, _clean2 = detect_cosmics(clean, readnoise=10., gain=1.0,
                                     sigclip=6, sigfrac=1.0, cleantype='medmask')
    assert _mask2.sum() == 0


def test_meanmask_clean(testdata):
    imdata, crmask = testdata
    # Because our image only contains single cosmics, turn off
    # neighbor detection. Also, our cosmic rays are high enough
    # contrast that we can turn our detection threshold up.
    _mask, clean = detect_cosmics(imdata, readnoise=10., gain=1.0,
                                  sigclip=6, sigfrac=1.0, cleantype='meanmask')

    assert (clean[crmask] != imdata[crmask]).sum() == crmask.sum()

    # Run it again on the clean data. We shouldn't find any new cosmic rays
    _mask2, _clean2 = detect_cosmics(clean, readnoise=10., gain=1.0,
                                     sigclip=6, sigfrac=1.0, cleantype='meanmask')
    assert _mask2.sum() == 0


def test_idw_clean(testdata):
    imdata, crmask = testdata
    # Because our image only contains single cosmics, turn off
    # neighbor detection. Also, our cosmic rays are high enough
    # contrast that we can turn our detection threshold up.
    _mask, clean = detect_cosmics(imdata, readnoise=10., gain=1.0,
                                  sigclip=6, sigfrac=1.0, cleantype='idw')

    assert (clean[crmask] != imdata[crmask]).sum() == crmask.sum()

    # Run it again on the clean data. We shouldn't find any new cosmic rays
    _mask2, _clean2 = detect_cosmics(clean, readnoise=10., gain=1.0,
                                     sigclip=6, sigfrac=1.0, cleantype='idw')
    assert _mask2.sum() == 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from numpy.testing import assert_allclose

from astroscrappy.utils import (median, optmed3, optmed5, optmed7, optmed9, optmed25,
                                medfilt3, medfilt5, medfilt7, sepmedfilt3, sepmedfilt5,
                                sepmedfilt7, sepmedfilt9, dilate3, dilate5, subsample,
                                rebin, laplaceconvolve, convolve)

import scipy.ndimage as ndi


def test_median():
    a = np.ascontiguousarray(np.random.random(1001)).astype('f4')
    assert np.float32(np.median(a)) == np.float32(median(a, 1001))


def test_optmed3():
    a = np.ascontiguousarray(np.random.random(3)).astype('f4')
    assert np.float32(np.median(a)) == np.float32(optmed3(a))


def test_optmed5():
    a = np.ascontiguousarray(np.random.random(5)).astype('f4')
    assert np.float32(np.median(a)) == np.float32(optmed5(a))


def test_optmed7():
    a = np.ascontiguousarray(np.random.random(7)).astype('f4')
    assert np.float32(np.median(a)) == np.float32(optmed7(a))


def test_optmed9():
    a = np.ascontiguousarray(np.random.random(9)).astype('f4')
    assert np.float32(np.median(a)) == np.float32(optmed9(a))


def test_optmed25():
    a = np.ascontiguousarray(np.random.random(25)).astype('f4')
    assert np.float32(np.median(a)) == np.float32(optmed25(a))


def test_medfilt3():
    a = np.ascontiguousarray(np.random.random((1001, 1001))).astype('f4')
    npmed3 = ndi.median_filter(a, size=(3, 3), mode='nearest')
    npmed3[:1, :] = a[:1, :]
    npmed3[-1:, :] = a[-1:, :]
    npmed3[:, :1] = a[:, :1]
    npmed3[:, -1:] = a[:, -1:]

    med3 = medfilt3(a)
    np.testing.assert_allclose(med3, npmed3)


def test_medfilt5():
    a = np.ascontiguousarray(np.random.random((1001, 1001))).astype('f4')
    npmed5 = ndi.median_filter(a, size=(5, 5), mode='nearest')
    npmed5[:2, :] = a[:2, :]
    npmed5[-2:, :] = a[-2:, :]
    npmed5[:, :2] = a[:, :2]
    npmed5[:, -2:] = a[:, -2:]

    med5 = medfilt5(a)
    np.testing.assert_allclose(med5, npmed5)


def test_medfilt7():
    a = np.ascontiguousarray(np.random.random((1001, 1001))).astype('f4')
    npmed7 = ndi.median_filter(a, size=(7, 7), mode='nearest')
    npmed7[:3, :] = a[:3, :]
    npmed7[-3:, :] = a[-3:, :]
    npmed7[:, :3] = a[:, :3]
    npmed7[:, -3:] = a[:, -3:]

    med7 = medfilt7(a)
    np.testing.assert_allclose(med7, npmed7)


def test_sepmedfilt3():
    a = np.ascontiguousarray(np.random.random((1001, 1001))).astype('f4')
    npmed3 = ndi.median_filter(a, size=(1, 3), mode='nearest')
    npmed3[:, :1] = a[:, :1]
    npmed3[:, -1:] = a[:, -1:]
    npmed3 = ndi.median_filter(npmed3, size=(3, 1), mode='nearest')
    npmed3[:1, :] = a[:1, :]
    npmed3[-1:, :] = a[-1:, :]
    npmed3[:, :1] = a[:, :1]
    npmed3[:, -1:] = a[:, -1:]

    med3 = sepmedfilt3(a)
    np.testing.assert_allclose(med3, npmed3)


def test_sepmedfilt5():
    a = np.ascontiguousarray(np.random.random((1001, 1001))).astype('f4')
    npmed5 = ndi.median_filter(a, size=(1, 5), mode='nearest')
    npmed5[:, :2] = a[:, :2]
    npmed5[:, -2:] = a[:, -2:]
    npmed5 = ndi.median_filter(npmed5, size=(5, 1), mode='nearest')
    npmed5[:2, :] = a[:2, :]
    npmed5[-2:, :] = a[-2:, :]
    npmed5[:, :2] = a[:, :2]
    npmed5[:, -2:] = a[:, -2:]

    med5 = sepmedfilt5(a)
    np.testing.assert_allclose(med5, npmed5)


def test_sepmedfilt7():
    a = np.ascontiguousarray(np.random.random((1001, 1001))).astype('f4')
    npmed7 = ndi.median_filter(a, size=(1, 7), mode='nearest')
    npmed7[:, :3] = a[:, :3]
    npmed7[:, -3:] = a[:, -3:]
    npmed7 = ndi.median_filter(npmed7, size=(7, 1), mode='nearest')
    npmed7[:3, :] = a[:3, :]
    npmed7[-3:, :] = a[-3:, :]
    npmed7[:, :3] = a[:, :3]
    npmed7[:, -3:] = a[:, -3:]

    med7 = sepmedfilt7(a)
    np.testing.assert_allclose(med7, npmed7)


def test_sepmedfilt9():
    a = np.ascontiguousarray(np.random.random((1001, 1001))).astype('f4')
    npmed9 = ndi.median_filter(a, size=(1, 9), mode='nearest')
    npmed9[:, :4] = a[:, :4]
    npmed9[:, -4:] = a[:, -4:]
    npmed9 = ndi.median_filter(npmed9, size=(9, 1), mode='nearest')
    npmed9[:4, :] = a[:4, :]
    npmed9[-4:, :] = a[-4:, :]
    npmed9[:, :4] = a[:, :4]
    npmed9[:, -4:] = a[:, -4:]

    med9 = sepmedfilt9(a)
    np.testing.assert_allclose(med9, npmed9)


def test_dilate5():
    niter = 2
    # Put 5% of the pixels into a mask
    a = np.zeros((1001, 1001), dtype=bool)
    np.random.seed(12431523)
    a[np.random.random((1001, 1001)) < 0.05] = True
    kernel = np.ones((5, 5))
    kernel[0, 0] = 0
    kernel[0, 4] = 0
    kernel[4, 0] = 0
    kernel[4, 4] = 0
    # Make a zero padded array for the numpy version to operate
    npdilate = ndi.binary_dilation(a, structure=kernel, iterations=1, border_value=0)

    def reset_edges(dilated, data):
        for i in range(2):
            dilated[:, i] = data[:, i]
            dilated[i, :] = data[i, :]
        for i in range(-2, 0, 1):
            dilated[:, i] = data[:, i]
            dilated[i, :] = data[i, :]

    reset_edges(npdilate, a)
    npdilate = ndi.binary_dilation(npdilate.copy(), structure=kernel, iterations=1, border_value=0)
    reset_edges(npdilate, a)
    cdilate = dilate5(a, niter)

    np.testing.assert_allclose(npdilate, cdilate)


def test_dilate3():
    # Put 5% of the pixels into a mask
    a = np.zeros((1001, 1001), dtype=bool)
    a[np.random.random((1001, 1001)) < 0.05] = True
    kernel = np.ones((3, 3))
    npgrow = ndi.binary_dilation(np.ascontiguousarray(a), structure=kernel)
    cgrow = dilate3(a)
    npgrow[:, 0] = a[:, 0]
    npgrow[:, -1] = a[:, -1]
    npgrow[0, :] = a[0, :]
    npgrow[-1, :] = a[-1, :]
    assert np.all(npgrow == cgrow)


def test_subsample():
    a = np.ascontiguousarray(np.random.random((1001, 1001))).astype('f4')
    npsubsamp = np.zeros((a.shape[0] * 2, a.shape[1] * 2), dtype=np.float32)
    for i in range(a.shape[0]):
        for j in range(a.shape[1]):
            npsubsamp[2 * i, 2 * j] = a[i, j]
            npsubsamp[2 * i + 1, 2 * j] = a[i, j]
            npsubsamp[2 * i, 2 * j + 1] = a[i, j]
            npsubsamp[2 * i + 1, 2 * j + 1] = a[i, j]

    csubsamp = subsample(a)
    assert np.all(npsubsamp == csubsamp)


def test_rebin():
    a = np.ascontiguousarray(np.random.random((2002, 2002)), dtype=np.float32)
    a = a.astype('f4')
    nprebin = np.zeros((1001, 1001), dtype=np.float32).astype('f4')
    for i in range(1001):
        for j in range(1001):
            nprebin[i, j] = (a[2 * i, 2 * j] + a[2 * i + 1, 2 * j] +
                             a[2 * i, 2 * j + 1] + a[2 * i + 1, 2 * j + 1])
            nprebin[i, j] /= np.float32(4.0)
    crebin = rebin(a)

    assert_allclose(crebin, nprebin, rtol=0, atol=1.e-6)


def test_laplaceconvolve():
    a = np.ascontiguousarray(np.random.random((1001, 1001))).astype('f4')
    k = np.array([[0.0, -1.0, 0.0], [-1.0, 4.0, -1.0], [0.0, -1.0, 0.0]])
    k = k.astype('<f4')
    npconv = ndi.convolve(a, k, mode='constant', cval=0.0)
    cconv = laplaceconvolve(a)
    assert_allclose(npconv, cconv, rtol=0.0, atol=1e-6)


def test_convolve():
    a = np.ascontiguousarray(np.random.random((1001, 1001))).astype('f4')
    k = np.ascontiguousarray(np.random.random((5, 5))).astype('f4')
    npconv = ndi.convolve(a, k, mode='constant', cval=0.0)
    cconv = convolve(a, k)
    assert_allclose(cconv, npconv, rtol=0, atol=1e-5)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from ..astroscrappy import detect_cosmics
from . import fake_data
import numpy as np


def test_main():
    imdata, expected_crmask = fake_data.make_fake_data()
    # Because our image only contains single cosmics, turn off
    # neighbor detection. Also, our cosmic rays are high enough
    # contrast that we can turn our detection threshold up.
    mask, _clean = detect_cosmics(imdata, readnoise=10., gain=1.0,
                                  sigclip=6, sigfrac=1.0)
    assert (mask == expected_crmask).sum() == (1001 * 1001)


def test_with_convolve_fine_structure():
    imdata, expected_crmask = fake_data.make_fake_data()
    # Convert from sigma to fwhm. Sigma is taken from the fake data utility.
    psf_fwhm = 3.5 * 2.0 * np.sqrt(2.0 * np.log(2.0))
    mask, _clean = detect_cosmics(imdata, readnoise=10., gain=1.0,
                                  sigclip=6, sigfrac=1.0, fsmode='convolve', psffwhm=psf_fwhm)
    assert (mask == expected_crmask).sum() == (1001 * 1001)



import os

from numpy import get_include
from setuptools import Extension

ROOT = os.path.relpath(os.path.dirname(__file__))


def get_extensions():
    sources = [os.path.join(ROOT, "astroscrappy.pyx")]
    ext = Extension(
        name="astroscrappy.astroscrappy",
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
        include_dirs=[get_include()],
        sources=sources,
    )
    return [ext]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os

import numpy as np
import scipy.ndimage as ndi
from astropy.io import fits

from ..astroscrappy import detect_cosmics

TESTFILE = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                        'data', 'gmos.fits')


def test_gmos():
    """This test uses a small cutout from a standard observation with GMOS
    (S20190808S0048). The extracted region is [362:, 480:680], and the file has
    been reduced with DRAGONS.
    """
    with fits.open(TESTFILE) as hdul:
        data = hdul['SCI'].data
        var = hdul['VAR'].data
        sky = hdul['SKYFIT'].data

    m1, _ = detect_cosmics(data, readnoise=4.24, gain=1.933)
    m2, _ = detect_cosmics(data, inbkg=sky, readnoise=4.24, gain=1.933)
    m3, _ = detect_cosmics(data, inbkg=sky, invar=var, readnoise=4.24, gain=1.933)

    cosmic1 = (slice(41, 72), slice(142, 161))
    cosmic2 = (slice(117, 147), slice(35, 43))

    # We must find 2 cosmic rays, but m1 (without bkg and var) also flags
    # 2 additional pixels that are identified as independent regions
    label, nb = ndi.label(m1)
    assert nb == 4
    objects = ndi.find_objects(label)
    assert cosmic1 in objects
    assert cosmic2 in objects
    areas = sorted([np.sum(label == (i+1)) for i in range(nb)])
    assert areas == [1, 1, 74, 93]

    for mask in m2, m3:
        label, nb = ndi.label(mask)
        assert nb == 2
        objects = ndi.find_objects(label)
        assert objects[0] == cosmic1
        assert objects[1] == cosmic2



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# cython: boundscheck=False, nonecheck=False, wraparound=False, language_level=3, cdivision=True
"""
Name : image_utils
Author : Curtis McCully
Date : October 2014
"""
import numpy as np
cimport numpy as np

np.import_array()

from libc.stdint cimport uint8_t


cdef extern from "imutils.h":
    void PySubsample(float * data, float * output, int nx, int ny) nogil
    void PyRebin(float * data, float * output, int nx, int ny) nogil
    void PyConvolve(float * data, float * kernel, float * output, int nx,
                    int ny, int kernx, int kerny) nogil
    void PyLaplaceConvolve(float * data, float * output, int nx, int ny) nogil
    void PyDilate3(uint8_t * data, uint8_t * output, int nx, int ny) nogil
    void PyDilate5(uint8_t * data, uint8_t * output, int niter, int nx,
                   int ny) nogil


def subsample(np.ndarray[np.float32_t, ndim=2, mode='c', cast=True] dsub):
    """subsample(dsub)\n
    Subsample an array 2x2 given an input array dsub.

    Parameters
    ----------
    dsub : float numpy array
        Array to be subsampled.

    Returns
    -------
    output : float numpy array
        Subsampled array. Output dimensions will be 2 times the input
        dimensions.

    Notes
    -----
    Each pixel is replicated into 4 pixels; no averaging is performed.
    The array needs to be C-contiguous order. Wrapper for PySubsample in
    imutils.
    """
    cdef int nx = dsub.shape[1]
    cdef int ny = dsub.shape[0]
    cdef int nx2 = 2 * nx
    cdef int ny2 = 2 * ny

    # Allocate the output array here so that Python tracks the memory and will
    # free the memory when we are finished with the output array.
    output = np.zeros((ny2, nx2), dtype=np.float32)

    cdef float * dsubptr = < float * > np.PyArray_DATA(dsub)
    cdef float * outdsubptr = < float * > np.PyArray_DATA(output)
    with nogil:
        PySubsample(dsubptr, outdsubptr, nx, ny)
    return output


def rebin(np.ndarray[np.float32_t, ndim=2, mode='c', cast=True] drebin):
    """rebin(drebin)\n
    Rebin an array 2x2.

    Rebin the array by block averaging 4 pixels back into 1.

    Parameters
    ----------
    drebin : float numpy array
        Array to be rebinned 2x2.

    Returns
    -------
    output : float numpy array
        Rebinned array. The size of the output array will be 2 times smaller
        than drebin.

    Notes
    -----
    This is effectively the opposite of subsample (although subsample does not
    do an average). The array needs to be C-contiguous order. Wrapper for
    PyRebin in imutils.
    """
    cdef int nx = drebin.shape[1] / 2
    cdef int ny = drebin.shape[0] / 2

    # Allocate the output array here so that Python tracks the memory and will
    # free the memory when we are finished with the output array.
    output = np.zeros((ny, nx), dtype=np.float32)

    cdef float * drebinptr = < float * > np.PyArray_DATA(drebin)
    cdef float * outdrebinptr = < float * > np.PyArray_DATA(output)
    with nogil:
        PyRebin(drebinptr, outdrebinptr, nx, ny)
    return output


def convolve(np.ndarray[np.float32_t, ndim=2, mode='c', cast=True] dconv,
             np.ndarray[np.float32_t, ndim=2, mode='c', cast=True] kernel):
    """convolve(dconv, kernel)\n
    Convolve an array with a kernel.

    Parameters
    ----------
    dconv : float numpy array
        Array to be convolved.

    kernel : float numpy array
        Kernel to use in the convolution.

    Returns
    -------
    output : float numpy array
        Convolved array.

    Notes
    -----
    Both the data and kernel arrays need to be C-contiguous order. Wrapper for
    PyConvolve in imutils.
    """
    cdef int nx = dconv.shape[1]
    cdef int ny = dconv.shape[0]

    # Allocate the output array here so that Python tracks the memory and will
    # free the memory when we are finished with the output array.
    output = np.zeros((ny, nx), dtype=np.float32)

    cdef float * dconvptr = < float * > np.PyArray_DATA(dconv)
    cdef float * outdconvptr = < float * > np.PyArray_DATA(output)

    cdef int knx = kernel.shape[1]
    cdef int kny = kernel.shape[0]
    cdef float * kernptr = < float * > np.PyArray_DATA(kernel)

    with nogil:
        PyConvolve(dconvptr, kernptr, outdconvptr, nx, ny, knx, kny)
    return output


def laplaceconvolve(np.ndarray[np.float32_t, ndim=2, mode='c', cast=True] dl):
    """laplaceconvolve(dl)\n
    Convolve an array with the Laplacian kernel.

    Convolve with the discrete version of the Laplacian operator with kernel:\n
     0 -1  0\n
    -1  4 -1\n
     0 -1  0\n

    Parameters
    ----------
    dl : float numpy array
        Array to be convolved.

    Returns
    -------
    output: float numpy array
        Convolved array.

    Notes
    -----
    The array needs to be C-contiguous order. Wrapper for PyLaplaceConvolve
    in imutils.
    """
    cdef int nx = dl.shape[1]
    cdef int ny = dl.shape[0]

    # Allocate the output array here so that Python tracks the memory and will
    # free the memory when we are finished with the output array.
    output = np.zeros((ny, nx), dtype=np.float32)

    cdef float * dlapptr = < float * > np.PyArray_DATA(dl)
    cdef float * outdlapptr = < float * > np.PyArray_DATA(output)
    with nogil:
        PyLaplaceConvolve(dlapptr, outdlapptr, nx, ny)
    return output


def dilate3(np.ndarray[np.uint8_t, ndim=2, mode='c', cast=True] dgrow):
    """dilate3(dgrow)\n
    Perform a boolean dilation on an array.

    Parameters
    ----------
    dgrow : boolean numpy array
        Array to dilate.

    Returns
    -------
    output : boolean numpy array
        Dilated array.

    Notes
    -----
    Dilation is the boolean equivalent of a convolution but using logical ors
    instead of a sum.
    We apply the following kernel:\n
    1 1 1\n
    1 1 1\n
    1 1 1\n
    The binary dilation is not computed for a 1 pixel border around the image.
    These pixels are copied from the input data. The array needs to be
    C-contiguous order. Wrapper for PyDilate3 in imutils.
    """
    cdef int nx = dgrow.shape[1]
    cdef int ny = dgrow.shape[0]

    # Allocate the output array here so that Python tracks the memory and will
    # free the memory when we are finished with the output array.
    output = np.zeros((ny, nx), dtype=bool)

    cdef uint8_t * dgrowptr = < uint8_t * > np.PyArray_DATA(dgrow)
    cdef uint8_t * outdgrowptr = < uint8_t * > np.PyArray_DATA(output)
    with nogil:
        PyDilate3(dgrowptr, outdgrowptr, nx, ny)
    return output


def dilate5(np.ndarray[np.uint8_t, ndim=2, mode='c', cast=True] ddilate,
            int niter):
    """dilate5(data, niter)\n
    Do niter iterations of boolean dilation on an array.

    Parameters
    ----------
    ddilate : boolean numpy array
        Array to dilate.

    niter : int
        Number of iterations.

    Returns
    -------
    output : boolean numpy array
        Dilated array.

    Notes
    -----
    Dilation is the boolean equivalent of a convolution but using logical ors
    instead of a sum.
    We apply the following kernel:\n
    0 1 1 1 0\n
    1 1 1 1 1\n
    1 1 1 1 1\n
    1 1 1 1 1\n
    0 1 1 1 0\n
    The edges are padded with zeros so that the dilation operator is defined
    for all pixels. The array needs to be C-contiguous order. Wrapper for
    PyDilate5 in imutils.
    """
    cdef int nx = ddilate.shape[1]
    cdef int ny = ddilate.shape[0]

    # Allocate the output array here so that Python tracks the memory and will
    # free the memory when we are finished with the output array.
    output = np.zeros((ny, nx), dtype=bool)

    cdef uint8_t * ddilateptr = < uint8_t * > np.PyArray_DATA(ddilate)
    cdef uint8_t * outddilateptr = < uint8_t * > np.PyArray_DATA(output)
    with nogil:
        PyDilate5(ddilateptr, outddilateptr, niter, nx, ny)
    return output



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# cython: boundscheck=False, nonecheck=False, wraparound=False, language_level=3, cdivision=True
# distutils: define_macros=NPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION
"""
Name : median_utils
Author : Curtis McCully
Date : October 2014
"""
import numpy as np
cimport numpy as np

np.import_array()


cdef extern from "medutils.h":
    float PyMedian(float * a, int n) nogil
    float PyOptMed3(float * a) nogil
    float PyOptMed5(float * a) nogil
    float PyOptMed7(float * a) nogil
    float PyOptMed9(float * a) nogil
    float PyOptMed25(float * a) nogil
    void PyMedFilt3(float * data, float * output, int nx, int ny) nogil
    void PyMedFilt5(float * data, float * output, int nx, int ny) nogil
    void PyMedFilt7(float * data, float * output, int nx, int ny) nogil
    void PySepMedFilt3(float * data, float * output, int nx, int ny) nogil
    void PySepMedFilt5(float * data, float * output, int nx, int ny) nogil
    void PySepMedFilt7(float * data, float * output, int nx, int ny) nogil
    void PySepMedFilt9(float * data, float * output, int nx, int ny) nogil


"""
Wrappers for the C functions in medutils.c
"""


def median(np.ndarray[np.float32_t, mode='c', cast=True] a, int n):
    """median(a, n)\n
    Find the median of the first n elements of an array.

    Parameters
    ----------
    a : float numpy array
        Input array to find the median.

    n : int
        Number of elements of the array to median.

    Returns
    -------
    med : float
        The median value.

    Notes
    -----
    Wrapper for PyMedian in medutils.
    """
    cdef float * aptr = < float * > np.PyArray_DATA(a)
    cdef float med = 0.0
    with nogil:
        med = PyMedian(aptr, n)
    return med

cdef float cymedian(float* a, int n) noexcept nogil:
    """cymedian(a, n)\n
    Cython function to calculate the median without requiring the GIL.
    :param a: float array to median
    :param n: int length of the array
    :return:
    """
    cdef float med = 0.0
    med = PyMedian(a, n)
    return med

def optmed3(np.ndarray[np.float32_t, ndim=1, mode='c', cast=True] a):
    """optmed3(a)\n
    Optimized method to find the median value of an array of length 3.

    Parameters
    ----------
    a : float numpy array
        Input array to find the median. Must be length 3.

    Returns
    -------
    med3 : float
        The median of the 3-element array.

    Notes
    -----
    Wrapper for PyOptMed3 in medutils.
    """
    cdef float * aptr3 = < float * > np.PyArray_DATA(a)
    cdef float med3 = 0.0
    with nogil:
        med3 = PyOptMed3(aptr3)
    return med3


def optmed5(np.ndarray[np.float32_t, ndim=1, mode='c', cast=True] a):
    """optmed5(a)\n
    Optimized method to find the median value of an array of length 5.

    Parameters
    ----------
    a : float numpy array
        Input array to find the median. Must be length 5.

    Returns
    -------
    med5 : float
        The median of the 5-element array.

    Notes
    -----
    Wrapper for PyOptMed5 in medutils.
    """
    cdef float * aptr5 = < float * > np.PyArray_DATA(a)
    cdef float med5 = 0.0
    with nogil:
        med5 = PyOptMed5(aptr5)
    return med5


def optmed7(np.ndarray[np.float32_t, ndim=1, mode='c', cast=True] a):
    """optmed7(a)\n
    Optimized method to find the median value of an array of length 7.

    Parameters
    ----------
    a : float numpy array
        Input array to find the median. Must be length 7.

    Returns
    -------
    med7 : float
        The median of the 7-element array.

    Notes
    -----
    Wrapper for PyOptMed7 in medutils.
    """
    cdef float * aptr7 = < float * > np.PyArray_DATA(a)
    cdef float med7 = 0.0
    with nogil:
        med7 = PyOptMed7(aptr7)
    return med7


def optmed9(np.ndarray[np.float32_t, ndim=1, mode='c', cast=True] a):
    """optmed9(a)\n
    Optimized method to find the median value of an array of length 9.

    Parameters
    ----------
    a : float numpy array
        Input array to find the median. Must be length 9.

    Returns
    -------
    med9 : float
        The median of the 9-element array.

    Notes
    -----
    Wrapper for PyOptMed9 in medutils.
    """
    cdef float * aptr9 = < float * > np.PyArray_DATA(a)
    cdef float med9 = 0.0
    with nogil:
        med9 = PyOptMed9(aptr9)
    return med9


def optmed25(np.ndarray[np.float32_t, ndim=1, mode='c', cast=True] a):
    """optmed25(a)\n
    Optimized method to find the median value of an array of length 25.

    Parameters
    ----------
    a : float numpy array
        Input array to find the median. Must be length 25.

    Returns
    -------
    med25 : float
        The median of the 25-element array.

    Notes
    -----
    Wrapper for PyOptMed25 in medutils.
    """
    cdef float * aptr25 = < float * > np.PyArray_DATA(a)
    cdef float med25 = 0.0
    with nogil:
        med25 = PyOptMed25(aptr25)
    return med25


def medfilt3(np.ndarray[np.float32_t, ndim=2, mode='c', cast=True] d3):
    """medfilt3(d3)\n
    Calculate the 3x3 median filter of an array.

    Parameters
    ----------
    d3 : float numpy array
        Array to median filter.

    Returns
    -------
    output : float numpy array
        Median filtered array.

    Notes
    -----
    The median filter is not calculated for a 1 pixel border around the image.
    These pixel values are copied from the input data. The array needs to be
    C-contiguous order. Wrapper for PyMedFilt3 in medutils.
    """
    cdef int nx = d3.shape[1]
    cdef int ny = d3.shape[0]

    # Allocate the output array here so that Python tracks the memory and will
    # free the memory when we are finished with the output array.
    output = np.zeros((ny, nx), dtype=np.float32)
    cdef float * d3ptr = < float * > np.PyArray_DATA(d3)
    cdef float * outd3ptr = < float * > np.PyArray_DATA(output)
    with nogil:
        PyMedFilt3(d3ptr, outd3ptr, nx, ny)

    return output


def medfilt5(np.ndarray[np.float32_t, ndim=2, mode='c', cast=True] d5):
    """medfilt5(d5)\n
    Calculate the 5x5 median filter of an array.

    Parameters
    ----------
    d5 : float numpy array
        Array to median filter.

    Returns
    -------
    output : float numpy array
        Median filtered array.

    Notes
    -----
    The median filter is not calculated for a 2 pixel border around the image.
    These pixel values are copied from the input data. The array needs to be
    C-contiguous order. Wrapper for PyMedFilt5 in medutils.
    """
    cdef int nx = d5.shape[1]
    cdef int ny = d5.shape[0]

    # Allocate the output array here so that Python tracks the memory and will
    # free the memory when we are finished with the output array.
    output = np.zeros((ny, nx), dtype=np.float32)
    cdef float * d5ptr = < float * > np.PyArray_DATA(d5)
    cdef float * outd5ptr = < float * > np.PyArray_DATA(output)
    with nogil:
        PyMedFilt5(d5ptr, outd5ptr, nx, ny)
    return output


def medfilt7(np.ndarray[np.float32_t, ndim=2, mode='c', cast=True] d7):
    """medfilt7(d7)\n
    Calculate the 7x7 median filter of an array.

    Parameters
    ----------
    d7 : float numpy array
        Array to median filter.

    Returns
    -------
    output : float numpy array
        Median filtered array.

    Notes
    -----
    The median filter is not calculated for a 3 pixel border around the image.
    These pixel values are copied from the input data. The array needs to be
    C-contiguous order. Wrapper for PyMedFilt7 in medutils.
    """
    cdef int nx = d7.shape[1]
    cdef int ny = d7.shape[0]

    # Allocate the output array here so that Python tracks the memory and will
    # free the memory when we are finished with the output array.
    output = np.zeros((ny, nx), dtype=np.float32)

    cdef float * d7ptr = < float * > np.PyArray_DATA(d7)
    cdef float * outd7ptr = < float * > np.PyArray_DATA(output)
    with nogil:
        PyMedFilt7(d7ptr, outd7ptr, nx, ny)
    return output


def sepmedfilt3(np.ndarray[np.float32_t, ndim=2, mode='c', cast=True] dsep3):
    """sepmedfilt3(dsep3)\n
    Calculate the 3x3 separable median filter of an array.

    Parameters
    ----------
    dsep3 : float numpy array
        Array to median filter.

    Returns
    -------
    output : float numpy array
        Median filtered array.

    Notes
    -----
    The separable median medians the rows followed by the columns instead of
    using a square window. Therefore it is not identical to the full median
    filter but it is approximately the same, but it is significantly faster.
    The median filter is not calculated for a 1 pixel border around the image.
    These pixel values are copied from the input data. The array needs to be
    C-contiguous order. Wrapper for PySepMedFilt3 in medutils.
    """
    cdef int nx = dsep3.shape[1]
    cdef int ny = dsep3.shape[0]

    # Allocate the output array here so that Python tracks the memory and will
    # free the memory when we are finished with the output array.
    output = np.zeros((ny, nx), dtype=np.float32)

    cdef float * dsep3ptr = < float * > np.PyArray_DATA(dsep3)
    cdef float * outdsep3ptr = < float * > np.PyArray_DATA(output)
    with nogil:
        PySepMedFilt3(dsep3ptr, outdsep3ptr, nx, ny)
    return np.asarray(output)


def sepmedfilt5(np.ndarray[np.float32_t, ndim=2, mode='c', cast=True] dsep5):
    """sepmedfilt5(dsep5)\n
    Calculate the 5x5 separable median filter of an array.

    Parameters
    ----------
    dsep5 : float numpy array
        Array to median filter.

    Returns
    -------
    output : float numpy array
        Median filtered array.

    Notes
    -----
    The separable median medians the rows followed by the columns instead of
    using a square window. Therefore it is not identical to the full median
    filter but it is approximately the same, but it is significantly faster.
    The median filter is not calculated for a 2 pixel border around the image.
    These pixel values are copied from the input data. The array needs to be
    C-contiguous order. Wrapper for PySepMedFilt5 in medutils.
    """
    cdef int nx = dsep5.shape[1]
    cdef int ny = dsep5.shape[0]

    # Allocate the output array here so that Python tracks the memory and will
    # free the memory when we are finished with the output array.
    output = np.zeros((ny, nx), dtype=np.float32)

    cdef float * dsep5ptr = < float * > np.PyArray_DATA(dsep5)
    cdef float * outdsep5ptr = < float * > np.PyArray_DATA(output)
    with nogil:
        PySepMedFilt5(dsep5ptr, outdsep5ptr, nx, ny)

    return output


def sepmedfilt7(np.ndarray[np.float32_t, ndim=2, mode='c', cast=True] dsep7):
    """sepmedfilt7(dsep7)\n
    Calculate the 7x7 separable median filter of an array.

    Parameters
    ----------
    dsep7 : float numpy array
        Array to median filter.

    Returns
    -------
    output : float numpy array
        Median filtered array.

    Notes
    -----
    The separable median medians the rows followed by the columns instead of
    using a square window. Therefore it is not identical to the full median
    filter but it is approximately the same, but it is significantly faster.
    The median filter is not calculated for a 3 pixel border around the image.
    These pixel values are copied from the input data. The array needs to be
    C-contiguous order. Wrapper for PySepMedFilt7 in medutils.
    """
    cdef int nx = dsep7.shape[1]
    cdef int ny = dsep7.shape[0]

    # Allocate the output array here so that Python tracks the memory and will
    # free the memory when we are finished with the output array.
    output = np.zeros((ny, nx), dtype=np.float32)

    cdef float * dsep7ptr = < float * > np.PyArray_DATA(dsep7)
    cdef float * outdsep7ptr = < float * > np.PyArray_DATA(output)
    with nogil:
        PySepMedFilt7(dsep7ptr, outdsep7ptr, nx, ny)
    return output


def sepmedfilt9(np.ndarray[np.float32_t, ndim=2, mode='c', cast=True] dsep9):
    """sepmedfilt9(dsep9)\n
    Calculate the 9x9 separable median filter of an array.

    Parameters
    ----------
    dsep9 : float numpy array
        Array to median filter.

    Returns
    -------
    output : float numpy array
        Median filtered array.

    Notes
    -----
    The separable median medians the rows followed by the columns instead of
    using a square window. Therefore it is not identical to the full median
    filter but it is approximately the same, but it is significantly faster.
    The median filter is not calculated for a 4 pixel border around the image.
    These pixel values are copied from the input data. The array needs to be
    C-contiguous order. Wrapper for PySepMedFilt9 in medutils.
    """

    cdef int nx = dsep9.shape[1]
    cdef int ny = dsep9.shape[0]

    # Allocate the output array here so that Python tracks the memory and will
    # free the memory when we are finished with the output array.
    output = np.zeros((ny, nx), dtype=np.float32)

    cdef float * dsep9ptr = < float * > np.PyArray_DATA(dsep9)
    cdef float * outdsep9ptr = < float * > np.PyArray_DATA(output)
    with nogil:
        PySepMedFilt9(dsep9ptr, outdsep9ptr, nx, ny)
    return output



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np


# Make a simple Gaussian function for testing purposes
def gaussian(image_shape, x0, y0, brightness, fwhm):
    x = np.arange(image_shape[1])
    y = np.arange(image_shape[0])
    x2d, y2d = np.meshgrid(x, y)

    sig = fwhm / 2.35482

    normfactor = brightness / 2.0 / np.pi * sig ** -2.0
    exponent = -0.5 * sig ** -2.0
    exponent *= (x2d - x0) ** 2.0 + (y2d - y0) ** 2.0

    return normfactor * np.exp(exponent)


def make_fake_data():
    """
    Generate fake data that can be used to test the detection and cleaning algorithms

    Returns
    -------
    imdata : numpy float array
        Fake Image data
    crmask : numpy boolean array
        Boolean mask of locations of injected cosmic rays
    """
    # Set a seed so that the tests are repeatable
    np.random.seed(200)

    # Create a simulated image to use in our tests
    imdata = np.zeros((1001, 1001), dtype=np.float32)

    # Add sky and sky noise
    imdata += 200

    psf_sigma = 3.5

    # Add some fake sources
    for i in range(100):
        x = np.random.uniform(low=0.0, high=1001)
        y = np.random.uniform(low=0.0, high=1001)
        brightness = np.random.uniform(low=1000., high=30000.)
        imdata += gaussian(imdata.shape, x, y, brightness, psf_sigma)

    # Add the poisson noise
    imdata = np.float32(np.random.poisson(imdata))

    # Add readnoise
    imdata += np.random.normal(0.0, 10.0, size=(1001, 1001))

    # Add 100 fake cosmic rays
    cr_x = np.random.randint(low=5, high=995, size=100)
    cr_y = np.random.randint(low=5, high=995, size=100)

    cr_brightnesses = np.random.uniform(low=1000.0, high=30000.0, size=100)

    imdata[cr_y, cr_x] += cr_brightnesses
    imdata = imdata.astype('f4')

    # Make a mask where the detected cosmic rays should be
    crmask = np.zeros((1001, 1001), dtype=bool)
    crmask[cr_y, cr_x] = True
    return imdata, crmask



# cython: language_level=3
"""
Header file for Cython functions in the utils package.

This allows the Cython code to call these routines directly
without requiring the GIL.
"""

"""
Calculate the median on the first n elements of C float array
without requiring the GIL.
"""
cdef float cymedian(float* aptr, int n) noexcept nogil



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# cython: boundscheck=False, nonecheck=False, wraparound=False, language_level=3, cdivision=True
"""
Name : astroscrappy: The Speedy Cosmic Ray Annihilation Package in Python
Author : Curtis McCully
Date : October 2014
"""

import numpy as np
cimport numpy as np

np.import_array()

cimport cython
from cython.parallel cimport parallel, prange

from .utils import *
from .utils.median_utils cimport cymedian

from libc.stdint cimport uint8_t

ctypedef uint8_t bool

from libc.stdlib cimport malloc, free

def detect_cosmics(indat, inmask=None, inbkg=None, invar=None, float sigclip=4.5,
                   float sigfrac=0.3, float objlim=5.0, float gain=1.0,
                   float readnoise=6.5, float satlevel=65536.0,
                   int niter=4, sepmed=True, cleantype='meanmask',
                   fsmode='median', psfmodel='gauss', float psffwhm=2.5,
                   int psfsize=7, psfk=None, float psfbeta=4.765,
                   verbose=False):
    """detect_cosmics(indat, inmask=None, inbkg=None, invar=None, sigclip=4.5,
                      sigfrac=0.3, objlim=5.0, gain=1.0, readnoise=6.5,
                      satlevel=65536.0, niter=4, sepmed=True,
                      cleantype='meanmask', fsmode='median', psfmodel='gauss',
                      psffwhm=2.5, psfsize=7, psfk=None, psfbeta=4.765,
                      verbose=False)\n
    Detect cosmic rays in a numpy array.

    If you use this code, please cite the Zendo DOI: https://zenodo.org/record/1482019

    Please cite the original paper which can be found at:
    http://www.astro.yale.edu/dokkum/lacosmic/

    van Dokkum 2001, PASP, 113, 789, 1420
    (article : http://adsabs.harvard.edu/abs/2001PASP..113.1420V)

    Parameters
    ----------
    indat : float numpy array
        Input data array that will be used for cosmic ray detection. This
        should include the sky background (or a mean background level, added
        back in after sky subtraction), so that noise can be estimated
        correctly from the data values. This should be in units of "counts".

    inmask : boolean numpy array, optional
        Input bad pixel mask. Values of True will be ignored in the cosmic ray
        detection/cleaning process. Default: None.

    inbkg : float numpy array, optional
        A pre-determined background image, to be subtracted from ``indat``
        before running the main detection algorithm.
        This is used primarily with spectroscopic data, to remove
        sky lines and the cross-section of an object continuum during
        iteration, "protecting" them from spurious rejection (see the above
        paper). This background is not removed from the final, cleaned output
        (`cleanarr`). This should be in units of "counts", the same units of indat.
        This inbkg should be free from cosmic rays. When estimating the cosmic-ray
        free noise of the image, we will treat ``inbkg`` as a constant Poisson
        contribution to the variance.

    invar : float numpy array, optional
        A pre-determined estimate of the data variance (ie. noise squared) in
        each pixel, generated by previous processing of ``indat``. If provided,
        this is used in place of an internal noise model based on ``indat``,
        ``gain`` and ``readnoise``. This still gets median filtered and cleaned
        internally, to estimate what the noise in each pixel *would* be in the
        absence of cosmic rays. This should be in units of "counts" squared.

    sigclip : float, optional
        Laplacian-to-noise limit for cosmic ray detection. Lower values will
        flag more pixels as cosmic rays. Default: 4.5.

    sigfrac : float, optional
        Fractional detection limit for neighboring pixels. For cosmic ray
        neighbor pixels, a lapacian-to-noise detection limit of
        sigfrac * sigclip will be used. Default: 0.3.

    objlim : float, optional
        Minimum contrast between Laplacian image and the fine structure image.
        Increase this value if cores of bright stars are flagged as cosmic
        rays. Default: 5.0.

    gain : float, optional
        Gain of the image (electrons / ADU). We always need to work in
        electrons for cosmic ray detection. Default: 1.0

    readnoise : float, optional
        Read noise of the image (electrons). Used to generate the noise model
        of the image. Default: 6.5.

    satlevel : float, optional
        Saturation of level of the image (electrons). This value is used to
        detect saturated stars and pixels at or above this level are added to
        the mask. Default: 65536.0.

    niter : int, optional
        Number of iterations of the LA Cosmic algorithm to perform. Default: 4.

    sepmed : boolean, optional
        Use the separable median filter instead of the full median filter.
        The separable median is not identical to the full median filter, but
        they are approximately the same and the separable median filter is
        significantly faster and still detects cosmic rays well. Default: True

    cleantype : {'median', 'medmask', 'meanmask', 'idw'}, optional
        Set which clean algorithm is used:\n
        'median': An umasked 5x5 median filter\n
        'medmask': A masked 5x5 median filter\n
        'meanmask': A masked 5x5 mean filter\n
        'idw': A masked 5x5 inverse distance weighted interpolation\n
        Default: "meanmask".

    fsmode : {'median', 'convolve'}, optional
        Method to build the fine structure image:\n
        'median': Use the median filter in the standard LA Cosmic algorithm
        'convolve': Convolve the image with the psf kernel to calculate the
        fine structure image using a matched filter technique.
        Default: 'median'.

    psfmodel : {'gauss', 'gaussx', 'gaussy', 'moffat'}, optional
        Model to use to generate the psf kernel if fsmode == 'convolve' and
        psfk is None. The current choices are Gaussian and Moffat profiles.
        'gauss' and 'moffat' produce circular PSF kernels. The 'gaussx' and
        'gaussy' produce Gaussian kernels in the x and y directions
        respectively. Default: "gauss".

    psffwhm : float, optional
        Full Width Half Maximum of the PSF to use to generate the kernel.
        Default: 2.5.

    psfsize : int, optional
        Size of the kernel to calculate. Returned kernel will have size
        psfsize x psfsize. psfsize should be odd. Default: 7.

    psfk : float numpy array, optional
        PSF kernel array to use for the fine structure image if
        fsmode == 'convolve'. If None and fsmode == 'convolve', we calculate
        the psf kernel using 'psfmodel'. Default: None.

    psfbeta : float, optional
        Moffat beta parameter. Only used if fsmode=='convolve' and
        psfmodel=='moffat'. Default: 4.765.

    verbose : boolean, optional
        Print to the screen or not. Default: False.

    Returns
    -------
    crmask : boolean numpy array
        The cosmic ray mask (boolean) array with values of True where there are
        cosmic ray detections.

    cleanarr : float numpy array
        The cleaned data array.

    Notes
    -----
    To reproduce the most similar behavior to the original LA Cosmic
    (written in IRAF), set  inmask = None, satlevel = np.inf, sepmed=False,
    cleantype='medmask', and fsmode='median'.

    The original IRAF version distinguishes between spectroscopic and imaging
    data. This version does not.

    For best results on spectra, we recommend that you include an estimate of the background.
    One can generally obtain this by fitting columns with a smooth function. To efficiently identify
    cosmic rays, LA Cosmic and therefore astroscrappy estimates the cosmic ray free noise by smoothing
    the variance using a median filter. To minimize false positives on bright sky lines, if ``inbkg`` is
    provided, we do not smooth the variance contribution from the provided background. We only smooth
    the variance that is in addition to the Poisson contribution from the background so that we do
    not underestimate the noise (and therefore run the risk of flagging false positives) near narrow, bright
    sky lines.
    """

    # Grab the sizes of the input array
    cdef int nx = indat.shape[1]
    cdef int ny = indat.shape[0]

    # Tell the compiler about the loop indices so it can optimize them.
    cdef int i, j = 0

    # Make a copy of the data as the cleanarr that we work on
    # This guarantees that that the data will be contiguous and makes sure we
    # don't edit the input data.
    cleanarr = np.empty((ny, nx), dtype=np.float32, order='C')
    # Set the initial values to those of the data array
    # Multiply by the gain; the statistics only work properly with electrons.
    cleanarr[:, :] = indat[:, :] * gain

    if inbkg is not None:
        bkg = np.empty((ny, nx), dtype=np.float32, order='C')
        bkg[:, :] = inbkg[:, :] * gain

    # Setup the mask
    if inmask is None:
        # By default don't mask anything
        mask = np.zeros((ny, nx), dtype=np.uint8, order='C')
    else:
        # Make a copy of the input mask
        mask = np.empty((ny, nx), dtype=np.uint8, order='C')
        mask[:, :] = inmask[:, :]

    # Copy any noise estimates so we can clean them like the data, otherwise
    # use the main data array instead.
    if invar is not None and inbkg is not None:
        clean_var = invar * (gain * gain) - bkg
    elif invar is not None and inbkg is None:
        clean_var = invar * (gain * gain)
    elif invar is None and inbkg is not None:
        clean_var = cleanarr - bkg + readnoise * readnoise
    else:
        clean_var = cleanarr + readnoise * readnoise

    # Find the saturated stars and add them to the mask
    update_mask(np.asarray(cleanarr), np.asarray(mask), satlevel, sepmed)

    # Subtract the input sky model, if applicable.
    if inbkg is not None:
        cleanarr -= bkg

    # Find the unmasked pixels to calculate the sky.
    gooddata = np.zeros(int(nx * ny - np.asarray(mask).sum()), dtype=np.float32,
                        order='c')

    igoodpix = 0

    gooddata[:] = cleanarr[np.logical_not(mask)]

    # Get the default background level for large cosmic rays.
    background_level = median(gooddata, len(gooddata))
    goodvar = np.empty_like(gooddata, order='c')
    goodvar[:] = clean_var[np.logical_not(mask)]
    background_var_level = median(goodvar, len(goodvar))
    del goodvar
    del gooddata

    # Set up the psf kernel if necessary.
    if psfk is None and fsmode == 'convolve':
        # calculate the psf kernel psfk
        if psfmodel == 'gauss':
            psfk = gausskernel(psffwhm, psfsize)
        elif psfmodel == 'gaussx':
            psfk = gaussxkernel(psffwhm, psfsize)
        elif psfmodel == 'gaussy':
            psfk = gaussykernel(psffwhm, psfsize)
        elif psfmodel == 'moffat':
            psfk = moffatkernel(psffwhm, psfbeta, psfsize)
        else:
            raise ValueError('Please choose a supported PSF model.')
        # Reverse the psf kernel as that is what we use in the convolution throughout
        psfk[:, :] = psfk[::-1, ::-1]

    # Define a cosmic ray mask
    # This is what will be returned at the end
    crmask = np.zeros((ny, nx), dtype=np.uint8, order='C')

    # Calculate the detection limit for neighbor pixels
    cdef float sigcliplow = sigfrac * sigclip

    # Run lacosmic for up to maxiter iterations
    # We stop if no more cosmic ray pixels are found (quite rare)
    if verbose:
        print("Starting {} L.A.Cosmic iterations".format(niter))
    for i in range(niter):
        if verbose:
            print("Iteration {}:".format(i + 1))

        # Detect the cosmic rays

        # We subsample, convolve, clip negative values,
        # and rebin to original size
        subsam = subsample(cleanarr)

        conved = laplaceconvolve(subsam)
        del subsam

        conved[conved < 0] = 0.0
        # This is called L+ in the original LA Cosmic/cosmics.py
        s = rebin(conved)
        del conved

        # Build a noise map, to compare the laplacian to, based on the input
        # data or (if supplied) input variance. If variance is used, it still
        # gets filtered, since we can best identify cosmic rays using what the
        # noise in a pixel *should* be, rather than what it actually is after
        # a cosmic ray hit.
        if sepmed:
            m5 = sepmedfilt7(clean_var)
        else:
            m5 = medfilt5(clean_var)

        # Here we add back in the background to get the full variance. m5 here is effectively the CR free variance
        if inbkg is not None:
            m5 += bkg
        # Clip noise so that we can take a square root
        m5[m5 < 0.00001] = 0.00001
        noise = np.sqrt(m5)
        del m5

        # Laplacian signal to noise ratio :
        s /= 2.0 * noise
        # the 2.0 is from the 2x2 subsampling and is denoted f_s in the original paper
        # This s is called sigmap in the original lacosmic.cl

        if sepmed:
            sp = sepmedfilt7(s)
        else:
            sp = medfilt5(s)

        # Remove the large structures (s prime) :
        sp = s - sp
        del s

        # Find the candidate cosmic rays
        goodpix = np.logical_not(mask)
        cosmics = np.logical_and(sp > sigclip, goodpix)

        # Build the fine structure image :
        # TODO: we can gain a reasonable performance boost if we only calculate the fine structure stuff for cosmic
        # pixels instead of the whole image
        if fsmode == 'convolve':
            # Use a match filter, the (reversed) psf convolved with the image / variance
            f = convolve(cleanarr / noise / noise, psfk)
            # Normalize by the sum of the weights to get the flux of the source assuming it was actually a psf.
            # This will pull the value of pixels that do not look like psfs (e.g. CRs) down.
            f = f / convolve(1.0 / noise / noise, psfk)
        elif fsmode == 'median':
            if sepmed:
                f = sepmedfilt5(cleanarr)
            else:
                f = medfilt3(cleanarr)
        else:
            raise ValueError('Please choose a valid fine structure mode.')

        if sepmed:
            m7 = sepmedfilt9(f)
        else:
            m7 = medfilt7(f)

        # Note that original paper compares L+ / Fine Structure Image < threshold.
        # The original code and therefore we compare S' / Fine Structure Image < threshold.
        # As such, we scale the fine structure image by the noise here so it is in the same units as S'.
        f = (f - m7) / noise
        # Clip f as we will divide by f. Similar to the IRAF version.
        f[f < 0.01] = 0.01

        del m7
        del noise

        # S' / F is still not exactly what is in the paper as we have subtracted the "sampling flux" from S
        # via the median filter. This should be an optimization because we have removed the smooth component
        # and are therefore comparing the candidate "CR flux" only.
        cosmics = np.logical_and(cosmics, (sp / f) > objlim)
        del f

        # What follows is a special treatment for neighbors, with more relaxed
        # constraints.
        # We grow these cosmics a first time to determine the immediate
        # neighborhood.
        cosmics = dilate3(cosmics)
        cosmics = np.logical_and(cosmics, goodpix)
        # From this grown set, we keep those that have sp > sigmaclip
        # Note that the only new pixels that should be added here are ones that were rejected by the
        # objlim test. As they are neighbors to true CRs, we take that as a prior that CR probability
        # of the neighboring pixel is enhanced.
        cosmics = np.logical_and(sp > sigclip, cosmics)

        # Now we repeat this procedure, but lower the detection limit to siglow
        cosmics = dilate3(cosmics)
        cosmics = np.logical_and(cosmics, goodpix)

        del goodpix
        cosmics = np.logical_and(sp > sigcliplow, cosmics)
        del sp

        # Our CR counter
        numcr = cosmics.sum()

        # Update the crmask with the cosmics we have found
        crmask[:, :] = np.logical_or(crmask, cosmics)[:, :]
        del cosmics
        if verbose:
            print("{} cosmic pixels this iteration".format(numcr))

        # If we didn't find anything, we're done.
        if numcr == 0:
            break

        # otherwise clean the image and iterate
        if cleantype == 'median':
            # Unmasked median filter
            clean_median(cleanarr, crmask, nx, ny)
            clean_median(clean_var, crmask, nx, ny)
        # Masked mean filter
        elif cleantype == 'meanmask':
            clean_meanmask(cleanarr, crmask, mask, nx, ny, background_level)
            clean_meanmask(clean_var, crmask, mask, nx, ny, background_var_level)
        # Masked median filter
        elif cleantype == 'medmask':
            clean_medmask(cleanarr, crmask, mask, nx, ny, background_level)
            clean_medmask(clean_var, crmask, mask, nx, ny, background_var_level)
        # Inverse distance weighted interpolation
        elif cleantype == 'idw':
            clean_idwinterp(cleanarr, crmask, mask, nx, ny, background_level)
            clean_idwinterp(clean_var, crmask, mask, nx, ny, background_var_level)
        else:
            raise ValueError("""cleantype must be one of the following values:
                            [median, meanmask, medmask, idw]""")

    # Restore the sky background if needed.
    if inbkg is not None:
        cleanarr += bkg

    cleanarr /= gain

    return crmask.astype(np.bool_), cleanarr


def update_mask(np.ndarray[np.float32_t, ndim=2, mode='c', cast=True] data,
                np.ndarray[np.uint8_t, ndim=2, mode='c', cast=True] mask,
                float satlevel, bool sepmed):
    """update_mask(data, mask, satlevel, sepmed)\n
     Find staturated stars and puts them in the mask.

     This can then be used to avoid these regions in cosmic detection and
     cleaning procedures. The median filter is used to find large symmetric
     regions of saturated pixels (i.e. saturated stars).

    Parameters
    ----------
    data : float numpy array
        The data array in which we look for saturated stars.

    mask : boolean numpy array
        Bad pixel mask. This mask will be dilated using dilate3 and then
        combined with the saturated star mask.

    satlevel : float
        Saturation level of the image. This value can be lowered if the cores
        of bright (saturated) stars are not being masked.

    sepmed : boolean
        Use the separable median or not. The separable median is not identical
        to the full median filter, but they are approximately the same and the
        separable median filter is significantly faster.
    """

    # Find all of the saturated pixels
    satpixels = data >= satlevel

    # Use the median filter to estimate the large scale structure
    if sepmed:
        m5 = sepmedfilt7(data)
    else:
        m5 = medfilt5(data)

    # Use the median filtered image to find the cores of saturated stars
    # The 10 here is arbitray. Malte Tewes uses 2.0 in cosmics.py, but I
    # wanted to get more of the cores of saturated stars.
    satpixels = np.logical_and(satpixels, m5 > (satlevel / 10.0))

    # Grow the input mask by one pixel to make sure we cover bad pixels
    grow_mask = dilate3(mask)

    # Dilate the saturated star mask to remove edge effects in the mask
    dilsatpixels = dilate5(satpixels, 2)
    del satpixels
    # Combine the saturated pixels with the given input mask
    # Note, we work on the mask pixels in place
    mask[:, :] = np.logical_or(dilsatpixels, grow_mask)[:, :]
    del grow_mask


cdef void clean_meanmask(float[:, ::1] cleanarr, bool[:, ::1] crmask,
                         bool[:, ::1] mask, int nx, int ny,
                         float background_level):
    """clean_meanmask(cleanarr, crmask, mask, nx, ny, background_level)\n
    Clean the bad pixels in cleanarr using a 5x5 masked mean filter.

    Parameters
    ----------
    cleanarr : float numpy array
        The array to be cleaned.

    crmask : boolean numpy array
        Cosmic ray mask. Pixels with a value of True in this mask will be
        cleaned.

    mask : boolean numpy array
        Bad pixel mask. Values of True indicate bad pixels.

    nx : int
        Size of cleanarr in the x-direction. Note cleanarr has dimensions
        ny x nx.

    ny : int
        Size of cleanarr in the y-direction. Note cleanarr has dimensions
        ny x nx.

    background_level : float
        Average value of the background. This value will be used if there are
        no good pixels in a 5x5 region.
    """

    # Go through all of the pixels, ignore the borders
    cdef int i, j, k, l, numpix
    cdef float s
    cdef bool badpix

    with nogil, parallel():
        # For each pixel
        for j in prange(2, ny - 2):
            for i in range(2, nx - 2):
                # if the pixel is in the crmask
                if crmask[j, i]:
                    numpix = 0
                    s = 0.0

                    # sum the 25 pixels around the pixel
                    # ignoring any pixels that are masked
                    for l in range(-2, 3):
                        for k in range(-2, 3):
                            badpix = crmask[j + l, i + k]
                            badpix = badpix or mask[j + l, i + k]
                            if not badpix:
                                s = s + cleanarr[j + l, i + k]
                                numpix = numpix + 1

                    # if the pixels count is 0
                    # then put in the background of the image
                    if numpix == 0:
                        s = background_level
                    else:
                        # else take the mean
                        s = s / float(numpix)

                    cleanarr[j, i] = s


cdef void clean_median(float[:, ::1] cleanarr, bool[:, ::1] crmask, int nx, int ny):
    """clean_medmask(cleanarr, crmask, mask, nx, ny, background_level)\n
    Clean the bad pixels in cleanarr using a 5x5 masked median filter.

    Parameters
    ----------
    cleanarr : float numpy array
        The array to be cleaned.

    crmask : boolean numpy array
        Cosmic ray mask. Pixels with a value of True in this mask will be
        cleaned.

    nx : int
        size of cleanarr in the x-direction. Note cleanarr has dimensions
        ny x nx.

    ny : int
        size of cleanarr in the y-direction. Note cleanarr has dimensions
        ny x nx.
    """
    # Go through all of the pixels, ignore the borders
    cdef int k, l, i, j, counter
    cdef float * medarr

    # For each pixel
    with nogil, parallel():
        medarr = < float * > malloc(25 * sizeof(float))
        for j in prange(2, ny - 2):
            for i in range(2, nx - 2):
                # if the pixel is in the crmask
                if crmask[j, i]:
                    # median the 25 pixels around the pixel
                    counter = 0
                    for l in range(-2, 3):
                        for k in range(-2, 3):
                            medarr[counter] = cleanarr[j + l, i + k]
                            counter =+ 1
                    cleanarr[j, i] =  cymedian(medarr, 25)

        free(medarr)


cdef void clean_medmask(float[:, ::1] cleanarr, bool[:, ::1] crmask,
                        bool[:, ::1] mask, int nx, int ny,
                        float background_level):
    """clean_medmask(cleanarr, crmask, mask, nx, ny, background_level)\n
    Clean the bad pixels in cleanarr using a 5x5 masked median filter.

    Parameters
    ----------
    cleanarr : float numpy array
        The array to be cleaned.

    crmask : boolean numpy array
        Cosmic ray mask. Pixels with a value of True in this mask will be
        cleaned.

    mask : boolean numpy array
        Bad pixel mask. Values of True indicate bad pixels.

    nx : int
        size of cleanarr in the x-direction. Note cleanarr has dimensions
        ny x nx.

    ny : int
        size of cleanarr in the y-direction. Note cleanarr has dimensions
        ny x nx.

    background_level : float
        Average value of the background. This value will be used if there are
        no good pixels in a 5x5 region.
    """
    # Go through all of the pixels, ignore the borders
    cdef int k, l, i, j, numpix
    cdef float * medarr
    cdef bool badpixel

    # For each pixel
    with nogil, parallel():
        medarr = < float * > malloc(25 * sizeof(float))
        for j in prange(2, ny - 2):
            for i in range(2, nx - 2):
                # if the pixel is in the crmask
                if crmask[j, i]:
                    numpix = 0
                    # median the 25 pixels around the pixel ignoring
                    # any pixels that are masked
                    for l in range(-2, 3):
                        for k in range(-2, 3):
                            badpixel = crmask[j + l, i + k]
                            badpixel = badpixel or mask[j + l, i + k]
                            if not badpixel:
                                medarr[numpix] = cleanarr[j + l, i + k]
                                numpix = numpix + 1

                    # if the pixels count is 0 then put in the background
                    # of the image
                    if numpix == 0:
                        cleanarr[j, i] = background_level
                    else:
                        # else take the median
                        cleanarr[j, i] =  cymedian(medarr, numpix)

        free(medarr)


cdef void clean_idwinterp(float[:, ::1] cleanarr, bool[:, ::1] crmask,
                          bool[:, ::1] mask, int nx, int ny,
                          float background_level):
    """clean_idwinterp(cleanarr, crmask, mask, nx, ny, background_level)\n
    Clean the bad pixels in cleanarr using a 5x5 using inverse distance
    weighted interpolation.

    Parameters
    ----------
    cleanarr : float numpy array
        The array to be cleaned.

    crmask : boolean numpy array
        Cosmic ray mask. Pixels with a value of True in this mask will be
        cleaned.

    mask : boolean numpy array
        Bad pixel mask. Values of True indicate bad pixels.

    nx : int
        Size of cleanarr in the x-direction (int). Note cleanarr has dimensions
        ny x nx.

    ny : int
        Size of cleanarr in the y-direction (int). Note cleanarr has dimensions
        ny x nx.

    background_level : float
        Average value of the background. This value will be used if there are
        no good pixels in a 5x5 region.
    """

    # Go through all the pixels, ignore the borders
    cdef int i, j, k, l
    cdef float f11, f12, f21, f22 = background_level
    cdef int x1, x2, y1, y2
    weightsarr = np.array([[0.35355339, 0.4472136, 0.5, 0.4472136, 0.35355339],
                          [0.4472136, 0.70710678, 1., 0.70710678, 0.4472136],
                          [0.5, 1., 0., 1., 0.5],
                          [0.4472136, 0.70710678, 1., 0.70710678, 0.4472136],
                          [0.35355339, 0.4472136, 0.5, 0.4472136, 0.35355339]],
                          dtype=np.float32)
    cdef float[:, ::1] weights = weightsarr
    cdef float wsum
    cdef float val
    cdef int x, y
    # For each pixel
    with nogil, parallel():

        for j in prange(2, ny - 2):
            for i in range(2, nx - 2):
                # if the pixel is in the crmask
                if crmask[j, i]:
                    wsum = 0.0
                    val = 0.0
                    for l in range(-2, 3):
                        y = j + l
                        for k in range(-2, 3):
                            x = i + k
                            if not (crmask[y, x] or mask[y, x]):
                                val = val + weights[l+2, k+2] * cleanarr[y, x]
                                wsum = wsum + weights[l+2, k+2]
                    if wsum < 1e-6:
                        cleanarr[j, i] = background_level
                    else:
                        cleanarr[j, i] = val / wsum


def gausskernel(float psffwhm, int kernsize):
    """gausskernel(psffwhm, kernsize)\n
    Calculate a circular Gaussian psf kernel.

    Parameters
    ----------
    psffwhm : float
        Full Width Half Maximum of the PSF to use to generate the kernel.

    kernsize : int
        Size of the kernel to calculate. kernsize should be odd.
        Returned kernel will have size kernsize x kernsize.

    Returns
    -------
    kernel : float numpy array
        Gaussian PSF kernel with size kernsize x kernsize.
    """
    kernel = np.zeros((kernsize, kernsize), dtype=np.float32)
    # Make a grid of x and y values
    x = np.tile(np.arange(kernsize) - kernsize / 2, (kernsize, 1))
    y = x.transpose().copy()
    # Calculate the offset, r
    r2 = x * x + y * y
    # Calculate the kernel
    sigma2 = psffwhm * psffwhm / 2.35482 / 2.35482
    kernel[:, :] = np.exp(-0.5 * r2 / sigma2)[:, :]
    # Normalize the kernel
    kernel /= kernel.sum()
    return kernel


def gaussxkernel(float psffwhm, int kernsize):
    """gaussxkernel(psffwhm, kernsize)\n
    Calculate a Gaussian kernel in the x-direction.

    This can be used for spectroscopic data.

    Parameters
    ----------
    psffwhm : float
        Full Width Half Maximum of the PSF to use to generate the kernel.

    kernsize : int
        Size of the kernel to calculate. kernsize should be odd.
        Returned kernel will have size kernsize x kernsize.

    Returns
    -------
    kernel : float numpy array
        Gaussian(x) kernel with size kernsize x kernsize.
    """
    kernel = np.zeros((kernsize, kernsize), dtype=np.float32)
    # Make a grid of x and y values
    x = np.tile(np.arange(kernsize) - kernsize / 2, (kernsize, 1))
    # Calculate the kernel
    sigma2 = psffwhm * psffwhm / 2.35482 / 2.35482
    kernel[:, :] = np.exp(-0.5 * x * x / sigma2)[:, :]
    # Normalize the kernel
    kernel /= kernel.sum()
    return kernel


def gaussykernel(float psffwhm, int kernsize):
    """gaussykernel(psffwhm, kernsize)\n
    Calculate a Gaussian kernel in the y-direction.

    This can be used for spectroscopic data.

    Parameters
    ----------
    psffwhm : float
        Full Width Half Maximum of the PSF to use to generate the kernel.

    kernsize : int
        Size of the kernel to calculate. kernsize should be odd.
        Returned kernel will have size kernsize x kernsize.

    Returns
    -------
    kernel : float numpy array
        Gaussian(y) kernel with size kernsize x kernsize.
    """
    kernel = np.zeros((kernsize, kernsize), dtype=np.float32)
    # Make a grid of x and y values
    x = np.tile(np.arange(kernsize) - kernsize / 2, (kernsize, 1))
    y = x.transpose().copy()
    # Calculate the kernel
    sigma2 = psffwhm * psffwhm / 2.35482 / 2.35482
    kernel[:, :] = np.exp(-0.5 * y * y / sigma2)[:, :]
    # Normalize the kernel
    kernel /= kernel.sum()
    return kernel


cdef moffatkernel(float psffwhm, float beta, int kernsize):
    """moffatkernel(psffwhm, beta, kernsize)\n
    Calculate a Moffat psf kernel.

    Parameters
    ----------
    psffwhm : float
        Full Width Half Maximum of the PSF to use to generate the kernel.

    beta : float
        Moffat beta parameter

    kernsize : int
        Size of the kernel to calculate. Returned kernel will have size
        kernsize x kernsize. kernsize should be odd.

    Returns
    -------
    kernel : float numpy array
        Moffat kernel with size kernsize x kernsize.
    """
    kernel = np.zeros((kernsize, kernsize), dtype=np.float32)
    # Make a grid of x and y values
    x = np.tile(np.arange(kernsize) - kernsize / 2, (kernsize, 1))
    y = x.transpose().copy()
    # Calculate the offset r
    r = np.sqrt(x * x + y * y)
    # Calculate the kernel
    hwhm = psffwhm / 2.0
    alpha = hwhm / np.sqrt(np.power(2.0, (1.0 / beta)) - 1.0)
    kernel[:, :] = (np.power(1.0 + (r * r / alpha / alpha), -1.0 * beta))[:, :]
    # Normalize the kernel.
    kernel /= kernel.sum()
    return kernel



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Astro-SCRAPPY: The Speedy Cosmic Ray Annihilation Package in Python
===================================================================

An optimized cosmic ray detector

:Author: Curtis McCully

Astro-SCRAPPY is designed to detect cosmic rays in images (numpy arrays),
originally based on Pieter van Dokkum's L.A.Cosmic algorithm.

Much of this was originally adapted from cosmics.py written by Malte Tewes.
I have ported all of the slow functions to Cython/C, and optimized
where I can. This is designed to be as fast as possible so some of the
readability has been sacrificed, specifically in the C code.

L.A.Cosmic = LAplacian Cosmic ray detection

If you use this code, please cite the Zendo DOI: https://zenodo.org/record/1482019

Please cite the original paper which can be found at:
http://www.astro.yale.edu/dokkum/lacosmic/

van Dokkum 2001, PASP, 113, 789, 1420
(article : http://adsabs.harvard.edu/abs/2001PASP..113.1420V)

This code requires Cython, preferably version >= 0.21.

Parallelization is achieved using OpenMP. This code should compile (although
the Cython files may have issues) using a compiler that does not support OMP,
e.g. clang.

Notes
-----
There are some differences from original LACosmic:

    - Automatic recognition of saturated stars.
      This avoids treating such stars as large cosmic rays.

    - I have tried to optimize all of the code as much as possible while
      maintaining the integrity of the algorithm. One of the key speedups is to
      use a separable median filter instead of the true median filter. While these
      are not identical, they produce comparable results and the separable version
      is much faster.

    - This implementation is much faster than the Python by as much as a factor of
      28 depending on the given parameters.
      This implementation is much faster than the original IRAF version, by a factor
      of ~90.

Note that arrays always must be C-contiguous, thus all loops are y outer, x inner.
This follows the Pyfits convention.

scipy is required for certain tests to pass, but the code itself does not depend on
scipy.
"""

from .astroscrappy import *  # noqa
from .utils import *  # noqa
from .version import version as __version__  # noqa

__all__ = ['detect_cosmics']  # noqa



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os

from Cython.Build import cythonize
from extension_helpers import get_extensions
from setuptools import setup

ext_modules = get_extensions()
compiler_directives = {}

if os.getenv('COVERAGE'):
    print('Adding linetrace directive')
    compiler_directives['profile'] = True
    compiler_directives['linetrace'] = True
    os.environ['CFLAGS'] = '-DCYTHON_TRACE_NOGIL=1 --coverage -fno-inline-functions -O0'

ext_modules = cythonize(ext_modules, compiler_directives=compiler_directives)

setup(ext_modules=ext_modules)



import os
import numpy as np

from setuptools import Extension
from extension_helpers import add_openmp_flags_if_available

UTIL_DIR = os.path.relpath(os.path.dirname(__file__))


def get_extensions():

    med_sources = [os.path.join(UTIL_DIR, "median_utils.pyx"),
                   os.path.join(UTIL_DIR, "medutils.c")]

    im_sources = [os.path.join(UTIL_DIR, "image_utils.pyx"),
                  os.path.join(UTIL_DIR, "imutils.c")]

    include_dirs = [np.get_include(), UTIL_DIR]

    if 'CFLAGS' in os.environ:
        extra_compile_args = os.environ['CFLAGS'].split()
    else:
        extra_compile_args = ['-g', '-O3', '-funroll-loops', '-ffast-math']

    ext_med = Extension(name='astroscrappy.utils.median_utils',
                        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
                        sources=med_sources,
                        include_dirs=include_dirs,
                        extra_compile_args=extra_compile_args)
    ext_im = Extension(name="astroscrappy.utils.image_utils",
                       define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
                       sources=im_sources,
                       include_dirs=include_dirs,
                       extra_compile_args=extra_compile_args)

    add_openmp_flags_if_available(ext_med)
    add_openmp_flags_if_available(ext_im)

    return [ext_med, ext_im]
